import torch
from torch import nn
from transformers import AutoTokenizer
from loguru import logger
from transformers.models.roberta import RobertaPreTrainedModel, RobertaModel
from typing import Optional
from base import Evaluator
from transformers import AdapterConfig
import transformers.adapters.composition as ac


def _truncate_seq_front(tokens, max_length):
    """Truncates a sequence pair in place to the maximum length."""
    while True:
        total_length = len(tokens)
        if total_length <= max_length:
            break
        else:
            tokens.pop(1)
        
class RobertaForPoEEvaluation(RobertaPreTrainedModel):
    def __init__(self, config):
        super().__init__(config)
        self.config = config
        self.roberta = RobertaModel(config, add_pooling_layer=False)
        self.names =   ["dailydialog", "empathetic", "topical", "reddit", "convai2"]
        
        
        self.heads = nn.ModuleDict(
            {
                name: nn.Linear(self.roberta.config.hidden_size, 1)
                for name in self.names
            }
        )
        
        logger.info(f"adding {len(self.names)} adapters")
        logger.info(f"adding adapters: {self.names}")
        
        config = AdapterConfig.load("houlsby")
        for name in self.names:
            self.roberta.add_adapter(name, config)
            self.roberta.train_adapter(name)
        
        self.sigmoid = nn.Sigmoid()
        # Initialize weights and apply final processing
        self.post_init()

    def forward(
        self,
        input_ids: Optional[torch.LongTensor] = None,
        attention_mask: Optional[torch.FloatTensor] = None,
    ):
        adapters = self.names
        self.roberta.active_adapters = ac.Parallel(*adapters)
        
        model_outputs = self.roberta(
            input_ids,
            attention_mask=attention_mask,
            return_dict=True
        )
        
        B, N = input_ids.shape
        K = len(adapters)
        hidden_states = model_outputs.last_hidden_state
        logit_lst = []
        # (B * K, 1) -> [(B, 1)] x K
        for adapter, hidden_state in zip(
                adapters, torch.mean(hidden_states, dim=1).split(B, dim=0)
        ):
            logit_lst.append(self.heads[adapter](hidden_state))

        # (B, K, 1)
        logits = torch.stack(logit_lst, dim=1)
        logits = torch.squeeze(self.sigmoid(logits), dim=-1)
        logger.info(logits)
        return logits


class PoeEvaluator(Evaluator):
    def __init__(self, args):
        super().__init__(args)
        self.tokenizer = AutoTokenizer.from_pretrained(args.poe_model_name_or_path, cache_dir='./cache')
        self.model = RobertaForPoEEvaluation.from_pretrained(args.poe_model_name_or_path, cache_dir='./cache').cuda()
        
    def get_name(self):
        return self.args.poe_model_name_or_path.split('/')[-1]       

    def get_score(self, context, response, reference=None, max_length=256):
        c = context.split('\n')
        c = ' '.join([utt[2:].strip() for utt in c])
        r = response
        a_input_ids = self.tokenizer.encode(c)
        b_input_ids = self.tokenizer.encode(r)
        input_ids = self.tokenizer.build_inputs_with_special_tokens(token_ids_0=a_input_ids[1:-1],
                                                                    token_ids_1=b_input_ids[1:-1])
        _truncate_seq_front(input_ids, max_length)
        input_masks = [1] * len(input_ids)
        while len(input_ids) < max_length:
            input_ids.append(self.tokenizer.pad_token_id)
            input_masks.append(0)

        model_inputs = dict()
        model_inputs['input_ids'] = torch.LongTensor([input_ids]).to(self.model.device)
        model_inputs['attention_mask'] = torch.LongTensor([input_masks]).to(self.model.device)
        logits = self.model(**model_inputs)
        scores = [1-s.item() for s in torch.squeeze(logits)]
        return scores
    
    @staticmethod
    def add_args(parser):
        parser.add_argument('--poe_model_name_or_path', default='zhangchen1991/poe-roberta-large')
        return parser